// $Id: quads-b_n.cpp,v 1.4 2006/02/04 13:15:56 maxhorn Exp $
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#define FO(i, n) for ((i) = 0; (i) < (n); ++(i))

typedef unsigned char byte;

#if defined(GF_9)
#include "gf9_table.h"
#elif defined(GF_25)
#include "gf25_table.h"
#elif defined(GF_49)
#include "gf49_table.h"
#elif defined(GF_81)
#include "gf81_table.h"
#endif

struct FieldElement {
	byte val;
	
    FieldElement() : val(0) {}

	FieldElement operator*(const FieldElement &b) const {
		FieldElement res(mulTable[val][b.val]);
		return res;
	}
	FieldElement operator/(const FieldElement &b) const {
		assert(b.val != 0);
		FieldElement res(mulTable[val][invTable[b.val]]);
		return res;
	}
	FieldElement operator+(const FieldElement &b) const {
		FieldElement res(addTable[val][b.val]);
		return res;
	}
	FieldElement operator-(const FieldElement &b) const {
		FieldElement res(addTable[val][negTable[b.val]]);
		return res;
	}

	FieldElement operator-() const {
		FieldElement res(negTable[val]);
		return res;
	}

	FieldElement &operator+=(const FieldElement &b) {
		val = addTable[val][b.val];
		return *this;
	}
	
	FieldElement &operator*=(const FieldElement &b) {
		val = mulTable[val][b.val];
		return *this;
	}
	
	FieldElement conj() const {
		FieldElement res(conjTable[val]);
		return res;
	}

	FieldElement norm() const {
		FieldElement res(mulTable[val][conjTable[val]]);
		return res;
	}

	static const FieldElement ZERO, ONE, TWO;

//protected:
	FieldElement(int i) {
		val = i % N;
	}
};

const FieldElement FieldElement::ZERO(0);
const FieldElement FieldElement::ONE(1);
const FieldElement FieldElement::TWO(addTable[1][1]);

typedef FieldElement OneDimSubspace[8];

static FieldElement prodA(const OneDimSubspace &u, const OneDimSubspace &v) {
	FieldElement sum(u[7] * v[7]);
	
	for(int i = 1; i <= 3; ++i) {
		sum += u[i] * v[3+i];
		sum += u[3+i] * v[i];
	}
	return sum;
}

static FieldElement prodB(const OneDimSubspace &u, const OneDimSubspace &v) {
	FieldElement sum;
	for (int i = 1; i <= 7; ++i)
		sum += u[i] * v[i].conj();
	
	return sum;
}

static inline bool IsGeometric(const OneDimSubspace &p) {
	return prodA(p, p).val == 0 && prodB(p, p).val != 0;
}

// The following functions assumes that p and q are isotropic w.r.t. prodA !
static bool IsGeometric(const OneDimSubspace &p, const OneDimSubspace &q) {
	if (prodA(p,q).val)
		return false;
	
	FieldElement nP = prodB(p,p);
	FieldElement nQ = prodB(q,q);
	FieldElement nPQ = prodB(p,q);
	
	return (nP * nQ - nPQ.norm()).val != 0;
}

// Check whether p is a point, and whether it splits the quadrangle a,b,c,d
static bool IsPointSplitting(
				const OneDimSubspace &a,
				const OneDimSubspace &b,
				const OneDimSubspace &c,
				const OneDimSubspace &d,
				const FieldElement &p3,
				const FieldElement &p6,
				const FieldElement &p7
				) {
	OneDimSubspace p;
	p[1] = -((c[6] + d[6]*b[1])*p3 +
			 (c[3] + d[3]*b[1])*p6 +
			 (c[7] + d[7]*b[1])*p7);
	p[2] = -(d[6]*p3 + d[3]*p6 + d[7]*p7);
	p[3] = p3;
	p[4] = 0;
	p[5] = 0;
	p[6] = p6;
	p[7] = p7;
 	return IsGeometric(p)
 		&& IsGeometric(p, a)
		&& IsGeometric(p, b)
		&& IsGeometric(p, c)
		&& IsGeometric(p, d)
		;
}

// Try to split the given quad a,b,c,d in the "obvious" way, i.e. by finding a
// point p that is collinear to all four corners.
static bool TryToSplitQuad(
				const OneDimSubspace &a,
				const OneDimSubspace &b,
				const OneDimSubspace &c,
				const OneDimSubspace &d
				) {
	FieldElement p7;
	
	// We need a geometric element which connects a,b,c,d.
	if (IsPointSplitting(a, b, c, d, 0, 1, 0)) 
		return true;
	for (p7.val = 0; p7.val < N; ++p7.val)
		if (IsPointSplitting(a, b, c, d, 1, -p7*p7/FieldElement::TWO, p7))
			return true;
	return false;
}

// Sometimes, TryToSplitQuad just is not good enough. In that case we try a
// bit harder: We scan through all points p that are collinear to a,c,d.
// Then, we check if the new quadrangle a,b,c,p can be split via TryToSplitQuad.
// If that works, we managed to decompose our original quad!
static bool TryToSplitQuadHard(
				const OneDimSubspace &a,
				const OneDimSubspace &b,
				const OneDimSubspace &c,
				const OneDimSubspace &d
				) {
	OneDimSubspace p;
	
	p[5].val = 1;
	for (p[3].val = 0; p[3].val <= 1; ++p[3].val) {
		for (p[6].val = 0; p[6].val < N; ++p[6].val) {
			for (p[7].val = 0; p[7].val < N; ++p[7].val) {
				p[2] = -(d[6]*p[3] + d[3]*p[6] + d[7]*p[7])
					+ p[5] * (d[3]*d[6] + d[7]*d[7]/FieldElement::TWO);
				p[1] = -(c[6]*p[3] + c[2]*p[5] + c[3]*p[6] + c[7]*p[7])
					+ p[2] * b[1];
				if (IsGeometric(p)
					&& IsGeometric(p, a)
					&& IsGeometric(p, c)
					&& IsGeometric(p, d)
					&& TryToSplitQuad(a, b, c, p))
					return true;
			}
		}
	}
	return false;
}

void printSubspace(const OneDimSubspace &a) {
	printf("(%d", a[1].val);
	for (int k = 2; k <= 7; ++k)
		printf(", %d", a[k].val);
	printf(")\n");

}

// The following program enumerates all relevant classes of quadrangles in
// the flip-flop geometry of SO(3,q). The "relevant" classes were determined
// with the help of computations performed in Mathematica. Essentially,
// the idea is to use the flag-transitive group action of SO(3,q) on the
// geometry in order to conjugate the vertices a and b such that a = e1 and 
//  b = lambda e1 + e2. Further restrictions on c and d are derived by using
// that (a,b)=(b,c)=(c,d)=(d,a) etc.

// For each quad, we then try to decompose it into triangles, using two methods
// implemented in TryToSplitQuad and TryToSplitQuad).

// The technique described above reduces us to q^16 relevants quads. This is
// far too much in the case (q=9) that interests us.

// First speedup trick: The number of orbits of SO(3,q) on GF(q^2)^3 is
// q^3+q^2+q (this includes the 0 orbit). We can use this to reduce the number
// of candidates for (c3,c6,c7) from q^6 to q^3+q^2+q.

// Furthermore, we can use that GU(2,q) acts on (e1,e2,f1,f2) to reduce the
// possible values of b[1] from q^2 to q-1.

// Consequently, instead of q^16 quadrangles, we only have to check a total
// of (q-1) * (q^3 + q^2 + q) * q^8 = q^12 - q^9 quads. Which is finally
// in the range we can handle, i.e. it "just" require a couple weeks on a 
// somewhat decent modern machine. And since the problem is trival to
// parallelize, we can do it in days or hours by simply using multiple
// machines.
int main(int argc, char **argv) {
	OneDimSubspace a, b, c, d;

	long long i = 0, j = 0, j2 = 0, jHard;

	// Change buffering to line buffered (not necessary for this program to
	// work correctly, it just makes it a bit nicer to use 'tee' to log the
	// output of it).
	setvbuf(stdout, 0, _IOLBF, 0);

	printf("Field size: %d\n", Q);
	
	int start_b1 = -1;
	int start_c2 = -1;
	int start_orb_rep = -1;
	
	// First param: the value for b1 to use. 
	if (argc > 1) {
		start_b1 = atoi(argv[1]);
		assert(0 <= start_b1 && start_b1 < Q);
	}

	// Second param: the value for c2 to use. 
	if (argc > 2) {
		start_c2 = atoi(argv[2]);
		assert(0 <= start_c2 && start_c2 < N);
	}

	// Third param: which orbit to use for c3/c6/c7
	if (argc > 3) {
		start_orb_rep = atoi(argv[3]);
		assert(0 <= start_orb_rep && start_orb_rep < Q*(Q*(Q+1)+1));
	}

	a[1].val = 1;
	b[2].val = 1;
	
	// Fr jeden der q mglichen Werte von x^(q+1), auer -1, mssen wir
	// fr b[1] nur jeweils einen Vertreter whlen.
	FieldElement normRepr[Q];
	for (i = N-1; i >= 0; --i) {
		FieldElement x(i);
		normRepr[x.norm().val] = x;
	}
	
	i = j = j2 = jHard = 0;
	for (int b1_norm = 0; b1_norm < Q; ++b1_norm) {
		// If a start value for b1 was specified, use it as an override
		if (start_b1 >= 0)
			b1_norm = start_b1;
		b[1] = normRepr[b1_norm];
		printf("Now scanning b[1] = %d\n", b[1].val);

		if (!IsGeometric(b)) {
			if (start_b1 >= 0)
				break;
			continue;
		}
		c[4].val = 1;
		c[5] = -b[1];
		
		FO(c[2].val, N) {
			// If start_c2 was specified, use it as an override
			if (start_c2 >= 0)
				c[2].val = start_c2;
		
			// We can restrict c3 and c6 to certain pairs of values.
			for (int orbRepIdx = 0; orbRepIdx < Q*(Q*(Q+1)+1); ++orbRepIdx) {
				// If start_orb_rep was specified, use it as an override
				if (start_orb_rep >= 0)
					orbRepIdx = start_orb_rep;
				c[3].val = orbReps[orbRepIdx][0];
				c[6].val = orbReps[orbRepIdx][1];
				c[7].val = orbReps[orbRepIdx][2];

				c[1] = c[2]*b[1] - c[3]*c[6] - c[7]*c[7]/FieldElement::TWO;


				if (!IsGeometric(c) ||
					!IsGeometric(b, c) ||
					prodA(a, c).val == 0) {
					if (start_orb_rep >= 0)
						break;
					continue;
				}
				
				printf("Now scanning: b1_norm %d, c2 %d, orb %d, total split %lld, cur split %lld, hard %lld\n",
						b1_norm, c[2].val, orbRepIdx, j, j2, jHard);

				i++;
				
				d[5].val = 1;
				j2 = 0;
				FO(d[3].val, N) {
					FO(d[6].val, N) {
						FO(d[7].val, N) {
						
							d[2] = -(d[3]*d[6] + d[7]*d[7]/FieldElement::TWO);
							d[1] = -(c[2] + c[6]*d[3] + c[3]*d[6] +
									 c[7]*d[7] - b[1]*d[2]);
							
							if (IsGeometric(d) && IsGeometric(c, d)
								&& IsGeometric(d,a) && prodA(b, d).val != 0) {
								bool success = TryToSplitQuad(a, b, c, d);

								if (!success) {
									success = TryToSplitQuadHard(a, b, c, d);
									jHard++;
								}
								
								j++;
								j2++;
								
								if (!success) {
									printf("Failed to split this quad:\n");
									printf("  a = "); printSubspace(a);
									printf("  b = "); printSubspace(b);
									printf("  c = "); printSubspace(c);
									printf("  d = "); printSubspace(d);
									exit(1);
								}
							}
						}
					}
				}

				if (start_orb_rep >= 0)
					break;
			}
			if (start_c2 >= 0)
				break;
		}
		
		// If a start value for b1 was specified, stop scanning now
		if (start_b1 >= 0)
			break;
	}

	printf("End stats: total split %lld, hard %lld\n", j, jHard);

	return 0;
}
